{% load i18n %}
<div style="text-align: center">
    <span class="step-links">
        {% if pages.number > 1 %}
            <a href="?{{qry_data}}&page=1">&lt; First</a>
        {% else %}
            &lt; {% translate "First" %}
        {% endif %}

        {% if pages.has_previous %}
            <a href="?{{qry_data}}&page={{ pages.previous_page_number }}">&lt;&lt; {% translate "Previous" %}</a>
        {% else %}
            &lt;&lt; {% translate "Previous" %}
        {% endif %}

        <span class="current" style="margin: 0 25px;">
            {% blocktrans with current=pages.number max=pages.paginator.num_pages %}Page {{ current }} of {{ max }}{% endblocktrans %}
        </span>

        {% if pages.has_next %}
            <a href="?{{qry_data}}&page={{ pages.next_page_number }}">{% translate "Next" %} &gt;&gt;</a>
        {% else %}
            {% translate "Next" %} &gt;&gt;
        {% endif %}

        {% if pages.number != pages.paginator.num_pages %}
            <a href="?{{qry_data}}&page={{ pages.paginator.num_pages }}">{% translate "Last" %} &gt;</a>
        {% else %}
            {% translate "Last" %} &gt;
        {% endif %}
    </span>
</div>
