{% extends "admin/change_list.html" %}
{% load i18n %}

{% block title %}
    {% translate "Link Checker" %} {{ block.super }}
{% endblock %}

{% block breadcrumbs %}
    <div class="breadcrumbs">
        <a href="{% url 'admin:index' %}">{% translate "Home" %}</a> &rsaquo;
        {% translate "Link Checker" %}
    </div>
{% endblock %}

{% block content %}
    <div id="content-main" class="nomargintop">
        <h1>{% translate "Link Checker" %}</h1>
        <div class="module">
            {% block innercontent %}
            {% endblock %}
        </div>
    </div>
{% endblock %}
