# Generated by Django 5.1.3 on 2024-11-25 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('linkcheck', '0010_url_add_error_message'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='link',
            index=models.Index(fields=['content_type', 'object_id'], name='content_type_and_object_id'),
        ),
    ]
